-- Isotropic membrane, Green strain

with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics.long_elementary_functions,
  terms,
  nlp;

procedure formfind is

  subtype fmat3x1 is terms.fmat1 (1 .. 3);

  subtype fmat9x1 is terms.fmat1 (1 .. 9);

  subtype fmat2x3 is terms.fmat2 (1 .. 2, 1 .. 3);

  type cablegroup_rec is
    record
      stiffness : long_float; -- Young's modulus * area
      density   : long_float; -- weight per unit length
    end record;

  type cablegroup_vec is array (integer range <>) of cablegroup_rec;

  type cablegroup_ptr is access cablegroup_vec;

  type cable_rec is
    record
      group  : integer;
      glob   : terms.imat1 (1 .. 2);
      length : long_float;
      u      : fmat3x1;
    end record;

  type cable_vec is array (integer range <>) of cable_rec;

  type cable_ptr is access cable_vec;

  type membrgroup_rec is
    record
      stiffness : long_float; -- Young's modulus * thickness
      poisson   : long_float; -- Poisson's ratio
      density   : long_float; -- weight per unit area
    end record;

  type membrgroup_vec is array (integer range <>) of membrgroup_rec;

  type membrgroup_ptr is access membrgroup_vec;

  type membr_rec is
    record
      group : integer;
      p     : long_float;
      area  : long_float;
      l1    : long_float;
      l2    : long_float;
      l3    : long_float;
      glob  : terms.imat1 (1 .. 3);
      u1    : terms.fmat1 (1 .. 3);
      u2    : terms.fmat1 (1 .. 3);
      u3    : terms.fmat1 (1 .. 3);
      h     : terms.fmat2 (1 .. 3, 1 .. 3);
    end record;

  type membr_vec is array (integer range <>) of membr_rec;

  type membr_ptr is access membr_vec;

  halt                        : exception;
  optimal                     : boolean := true;
  minor                       : integer := 0;
  nn,n,maxls,maxminor,maxlist : integer;
  major,maxmajor              : integer;
  lstol,mntol,mnref,maxnorm,f : long_float;
  cablegroup                  : cablegroup_ptr;
  membrgroup                  : membrgroup_ptr;
  index                       : terms.imat2_ptr;
  xo,yo,zo,x,g                : terms.fmat1_ptr;
  known,action,force          : terms.fmat2_ptr;
  cable                       : cable_ptr;
  membr                       : membr_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  procedure putchar (io : in text_io.file_type;
                     c  : in character;
                     n  : in integer) is

  begin
    for k in 1 .. n loop
      text_io.put(io,c);
    end loop;
  end putchar;

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    e,glob,axis,nk : integer;
    io             : file_type;

  begin
    open(io, in_file, "input.txt");

    -- solution parameters

    -- The iterations terminate if the infinity norm of the
    -- gradient becomes less than or equal to MAXNORM, which is
    -- calculated as:

    -- MAXNORM = MNTOL * MNREF

    -- If MNTOL is specified as a non-positive value, it is
    -- replaced by 0.001.

    -- If MNREF is specified as a non-positive value, it is
    -- replaced by the infinite norm of the gradient, evaluated at
    -- the starting point.

    -- If MAXMINOR is specified as a non-positive value, it is
    -- replaced by ten times the number of degrees of freedom.

    scan(io);
    get(io,lstol);
    get(io,maxls);
    get(io,mntol);
    get(io,mnref);
    get(io,maxminor);
    get(io,maxlist);
    get(io,maxmajor);
    skip_line(io);

    -- nodes
    scan(io);
    get(io,nn);
    skip_line(io);

    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    skip_line(io);
    for k in 1 .. nn loop
      get(io,glob);
      get(io,xo(glob));
      get(io,yo(glob));
      get(io,zo(glob));
      skip_line(io);
    end loop;

    -- cable groups
    scan(io);
    get(io,nk);
    skip_line(io);

    cablegroup := new cablegroup_vec (1 .. nk);

    if nk > 0 then
      skip_line(io);

      for k in 1 .. nk loop
        get(io,e);
        get(io,cablegroup(e).stiffness);
        get(io,cablegroup(e).density);
        skip_line(io);
      end loop;
    end if;

    -- cable elements
    scan(io);
    get(io,nk);
    skip_line(io);

    cable := new cable_vec (1 .. nk);

    if nk > 0 then
      skip_line(io);

      for k in 1 .. nk loop
        get(io,e);
        get(io,cable(e).glob(1));
        get(io,cable(e).glob(2));
        get(io,cable(e).group);
        skip_line(io);
      end loop;
    end if;

    -- membrane groups
    scan(io);
    get(io,nk);
    skip_line(io);

    membrgroup := new membrgroup_vec (1 .. nk);

    if nk > 0 then
      skip_line(io);

      for k in 1 .. nk loop
        get(io,e);
        get(io,membrgroup(e).stiffness);
        get(io,membrgroup(e).poisson);
        get(io,membrgroup(e).density);
        skip_line(io);
      end loop;
    end if;

    -- membrane elements
    scan(io);
    get(io,nk);
    skip_line(io);

    membr := new membr_vec (1 .. nk);

    if nk > 0 then
      skip_line(io);

      for k in 1 .. nk loop
        get(io,e);
        get(io,membr(e).glob(1));
        get(io,membr(e).glob(2));
        get(io,membr(e).glob(3));
        get(io,membr(e).group);
        skip_line(io);
      end loop;
    end if;

    -- initiate index
    index := new terms.imat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        index(glob,axis) := 0;
      end loop;
    end loop;

    -- initiate known
    known := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        known(glob,axis) := 0.0;
      end loop;
    end loop;

    -- known displacements
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,known(glob,axis));
        skip_line(io);
        index(glob,axis) := 1;
      end loop;
    end if;

    -- initiate action
    action := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        action(glob,axis) := 0.0;
      end loop;
    end loop;

    -- node actions
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,glob);
        get(io,axis);
        get(io,action(glob,axis));
        skip_line(io);
      end loop;
    end if;

    -- pressure on membrane elements
    scan(io);
    get(io,nk);
    skip_line(io);

    if nk > 0 then
      skip_line(io);
      for k in 1 .. nk loop
        get(io,e);
        get(io,membr(e).p);
        skip_line(io);
      end loop;
    end if;

    close(io);
  end input;

  procedure vertex (glob : in integer;
                    vx   : out fmat3x1) is

    i : integer;

  begin
    vx(1) := xo(glob);
    vx(2) := yo(glob);
    vx(3) := zo(glob);

    for axis in 1 .. 3 loop
      i := index(glob,axis);
      if i <= n then
        vx(axis) := vx(axis) + x(i);
      else
        vx(axis) := vx(axis) + known(glob,axis);
      end if;
    end loop;
  end vertex;

  procedure setforce is

    glob        : integer;
    aux         : long_float;
    vx1,vx2,vx3 : fmat3x1;
    u,v,w       : terms.fmat1 (1 .. 3);

  begin
    -- initiate force
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := action(glob,axis);
      end loop;
    end loop;

    -- add pressure
    for e in membr'range loop
      glob := membr(e).glob(1);
      vertex(glob,vx1);

      glob := membr(e).glob(2);
      vertex(glob,vx2);

      glob := membr(e).glob(3);
      vertex(glob,vx3);

      -- u, v
      for axis in 1 .. 3 loop
        u(axis) := vx3(axis) - vx2(axis);
        v(axis) := vx1(axis) - vx3(axis);
      end loop;

      -- w = u x v
      w(1) := u(2) * v(3) - u(3) * v(2);
      w(2) := u(3) * v(1) - u(1) * v(3);
      w(3) := u(1) * v(2) - u(2) * v(1);

      -- pressure
      aux := membr(e).p / 6.0;
      for axis in 1 .. 3 loop
        w(axis) := aux * w(axis);
      end loop;

      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          force(glob,axis) := force(glob,axis) + w(axis);
        end loop;
      end loop;
    end loop;
  end setforce;

  procedure cableobj (cable           : in cable_rec;
                      strain,axial,fe : out long_float;
                      x               : in terms.fmat1_ptr;
                      gc              : out fmat2x3) is

    use
      ada.numerics.long_elementary_functions;

    glob,i        : integer;
    c,alpha,normw : long_float;
    z,w           : terms.fmat1 (1 .. 3);
    xe            : terms.fmat2 (1 .. 2, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 2 loop
      glob := cable.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / cable.length;
    end loop;

    -- w
    normw := 0.0;
    for axis in 1 .. 3 loop
      w(axis) := cable.u(axis) + z(axis);
      normw := normw + w(axis) ** 2;
    end loop;
    normw := sqrt(normw);

    -- alpha
    alpha := 0.0;
    for axis in 1 .. 3 loop
      alpha := alpha + z(axis) * (cable.u(axis) + w(axis));
    end loop;

    -- strain
    strain := alpha / (normw + 1.0);

    -- axial
    axial := cablegroup(cable.group).stiffness * strain;

    -- energy
    fe := 0.5 * axial * strain * cable.length;

    -- residue
    c := axial / normw;
    for axis in 1 .. 3 loop
      alpha := c * w(axis);
      gc(1,axis) := -alpha;
      gc(2,axis) :=  alpha;
    end loop;
  end cableobj;

  procedure membrobj (membr          : in membr_rec;
                      st1,st2,st3,fe : out long_float;
                      x              : in terms.fmat1_ptr;
                      gm             : out fmat9x1) is

    glob,i                      : integer;
    delta1,e1,e2,e3,sv1,sv2,sv3 : long_float;
    z                           : terms.fmat1 (1 .. 3);
    de1,de2,de3                 : terms.fmat1 (1 .. 9);
    xe                          : terms.fmat2 (1 .. 3, 1 .. 3);

  begin
    -- global displacements
    for loc in 1 .. 3 loop
      glob := membr.glob(loc);
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i > n then
          xe(loc,axis) := known(glob,axis);
        else
          xe(loc,axis) := x(i);
        end if;
      end loop;
    end loop;

    -- side 1

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(3,axis) - xe(2,axis)) / membr.l1;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (membr.u1(axis) + z(axis)) ** 2;
    end loop;

    -- e1
    e1 := (delta1 - 1.0) / 2.0;

    -- de1
    de1(1) := 0.0;
    de1(2) := 0.0;
    de1(3) := 0.0;

    de1(7) := (membr.u1(1) + z(1)) / membr.l1;
    de1(8) := (membr.u1(2) + z(2)) / membr.l1;
    de1(9) := (membr.u1(3) + z(3)) / membr.l1;

    de1(4) := -de1(7);
    de1(5) := -de1(8);
    de1(6) := -de1(9);

    -- side 2

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(1,axis) - xe(3,axis)) / membr.l2;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (membr.u2(axis) + z(axis)) ** 2;
    end loop;

    -- e2
    e2 := (delta1 - 1.0) / 2.0;

    -- de2
    de2(4) := 0.0;
    de2(5) := 0.0;
    de2(6) := 0.0;

    de2(1) := (membr.u2(1) + z(1)) / membr.l2;
    de2(2) := (membr.u2(2) + z(2)) / membr.l2;
    de2(3) := (membr.u2(3) + z(3)) / membr.l2;

    de2(7) := -de2(1);
    de2(8) := -de2(2);
    de2(9) := -de2(3);

    -- side 3

    -- z
    for axis in 1 .. 3 loop
      z(axis) := (xe(2,axis) - xe(1,axis)) / membr.l3;
    end loop;

    -- delta + 1.0
    delta1 := 0.0;
    for axis in 1 .. 3 loop
      delta1 := delta1 + (membr.u3(axis) + z(axis)) ** 2;
    end loop;

    -- e3
    e3 := (delta1 - 1.0) / 2.0;

    -- de3
    de3(7) := 0.0;
    de3(8) := 0.0;
    de3(9) := 0.0;

    de3(4) := (membr.u3(1) + z(1)) / membr.l3;
    de3(5) := (membr.u3(2) + z(2)) / membr.l3;
    de3(6) := (membr.u3(3) + z(3)) / membr.l3;

    de3(1) := -de3(4);
    de3(2) := -de3(5);
    de3(3) := -de3(6);

    -- stress * thickness
    st1 := membr.h(1,1) * e1 + membr.h(1,2) * e2 + membr.h(1,3) * e3;
    st2 := membr.h(2,1) * e1 + membr.h(2,2) * e2 + membr.h(2,3) * e3;
    st3 := membr.h(3,1) * e1 + membr.h(3,2) * e2 + membr.h(3,3) * e3;

    -- stress * volume
    sv1 := st1 * membr.area;
    sv2 := st2 * membr.area;
    sv3 := st3 * membr.area;

    -- residue
    for i in 1 .. 9 loop
      gm(i) := sv1 * de1(i) + sv2 * de2(i) + sv3 * de3(i);
    end loop;

    -- energy
    fe := (sv1 * e1 + sv2 * e2 + sv3 * e3) / 2.0;
  end membrobj;

  procedure objective (n : in integer;
                       f : out long_float;
                       x : in terms.fmat1_ptr;
                       g : in terms.fmat1_ptr) is

    glob,i,s                    : integer;
    strain,axial,st1,st2,st3,fe : long_float;
    gc                          : fmat2x3;
    gm                          : fmat9x1;

  begin
    -- initiate objective, gradient
    f := 0.0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        i := index(glob,axis);
        if i <= n then
          f := f - force(glob,axis) * x(i);
          g(i) := -force(glob,axis);
        end if;
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      cableobj(cable(e),strain,axial,fe,x,gc);

      -- update objective
      f := f + fe;

      -- update gradient
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gc(loc,axis);
          end if;
        end loop;
      end loop;
    end loop;

    -- membrane elements
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gm);

      -- update objective
      f := f + fe;

      -- update gradient
      s := 0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          i := index(glob,axis);
          if i <= n then
            g(i) := g(i) + gm(s);
          end if;
        end loop;
      end loop;
    end loop;
  end objective;

  package qn is new nlp.qnewton(objective);

  procedure setup is

    use
      ada.numerics.long_elementary_functions;

    glob1,glob2,glob3,s1,s2,glob  : integer;
    h00,norm,poisson,weight       : long_float;
    cos1,cos2,cos3,sin1,sin2,sin3 : long_float;
    u                             : terms.fmat1 (1 .. 3);

  begin
    -- number of unknowns
    n := 0;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          n := n + 1;
        end if;
      end loop;
    end loop;

    -- new sequence
    s1 := 0;
    s2 := n;
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        if index(glob,axis) = 0 then
          s1 := s1 + 1;
          index(glob,axis) := s1;
        else
          s2 := s2 + 1;
          index(glob,axis) := s2;
        end if;
      end loop;
    end loop;

    -- cable elements
    for e in cable'range loop
      glob1 := cable(e).glob(1);
      glob2 := cable(e).glob(2);

      u(1) := xo(glob2) - xo(glob1);
      u(2) := yo(glob2) - yo(glob1);
      u(3) := zo(glob2) - zo(glob1);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);

      cable(e).length := norm;
      cable(e).u(1) := u(1) / norm;
      cable(e).u(2) := u(2) / norm;
      cable(e).u(3) := u(3) / norm;

      -- self weight
      weight := cable(e).length * cablegroup(cable(e).group).density / 2.0;
      glob := cable(e).glob(1);
      action(glob,3) := action(glob,3) - weight;
      glob := cable(e).glob(2);
      action(glob,3) := action(glob,3) - weight;
    end loop;

    -- membrane elements
    for e in membr'range loop
      glob1 := membr(e).glob(1);
      glob2 := membr(e).glob(2);
      glob3 := membr(e).glob(3);

      -- u1
      u(1) := xo(glob3) - xo(glob2);
      u(2) := yo(glob3) - yo(glob2);
      u(3) := zo(glob3) - zo(glob2);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        membr(e).u1(axis) := u(axis) / norm;
      end loop;
      membr(e).l1 := norm;

      -- u2
      u(1) := xo(glob1) - xo(glob3);
      u(2) := yo(glob1) - yo(glob3);
      u(3) := zo(glob1) - zo(glob3);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        membr(e).u2(axis) := u(axis) / norm;
      end loop;
      membr(e).l2 := norm;

      -- u3
      u(1) := xo(glob2) - xo(glob1);
      u(2) := yo(glob2) - yo(glob1);
      u(3) := zo(glob2) - zo(glob1);
      norm := sqrt(u(1) ** 2 + u(2) ** 2 + u(3) ** 2);
      for axis in 1 .. 3 loop
        membr(e).u3(axis) := u(axis) / norm;
      end loop;
      membr(e).l3 := norm;

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - membr(e).u2(axis) * membr(e).u3(axis);
      end loop;

      if cos1 > 1.0 then
        cos1 := 1.0;
      end if;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - membr(e).u3(axis) * membr(e).u1(axis);
      end loop;

      if cos2 > 1.0 then
        cos2 := 1.0;
      end if;

      -- sine alpha1
      sin1 := sqrt(1.0 - cos1 ** 2);

      -- sine alpha2
      sin2 := sqrt(1.0 - cos2 ** 2);

      -- cosine alpha3
      cos3 := sin1 * sin2 - cos1 * cos2;

      -- sine alpha3
      sin3 := sin1 * cos2 + sin2 * cos1;

      -- matrix H_
      poisson := membrgroup(membr(e).group).poisson;
      h00 := membrgroup(membr(e).group).stiffness / (1.0 - poisson ** 2) / 2.0;

      -- H
      membr(e).h(1,1) := h00 * (cos1 ** 2 + 1.0 - poisson * sin1 ** 2) / sin2 ** 2 / sin3 ** 2;
      membr(e).h(1,2) := h00 * (cos1 * cos2 - cos3 + poisson * sin1 * sin2) / sin1 / sin2 / sin3 ** 2;
      membr(e).h(1,3) := h00 * (cos1 * cos3 - cos2 + poisson * sin1 * sin3) / sin1 / sin2 ** 2 / sin3;
      membr(e).h(2,2) := h00 * (cos2 ** 2 + 1.0 - poisson * sin2 ** 2) / sin1 ** 2 / sin3 ** 2;
      membr(e).h(2,3) := h00 * (cos2 * cos3 - cos1 + poisson * sin2 * sin3) / sin1 ** 2 / sin2 / sin3;
      membr(e).h(3,3) := h00 * (cos3 ** 2 + 1.0 - poisson * sin3 ** 2) / sin1 ** 2 / sin2 ** 2;
      -- symmetry
      membr(e).h(2,1) := membr(e).h(1,2);
      membr(e).h(3,1) := membr(e).h(1,3);
      membr(e).h(3,2) := membr(e).h(2,3);

      -- area
      membr(e).area := membr(e).l1 * membr(e).l2 * sin3 / 2.0;

      -- self weight
      weight := membr(e).area * membrgroup(membr(e).group).density / 3.0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        action(glob,3) := action(glob,3) - weight;
      end loop;
    end loop;

    -- initiate x
    x := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      x(k) := 0.0;
    end loop;

    -- initiate g
    g := new terms.fmat1 (1 .. n);
    for k in 1 .. n loop
      g(k) := 0.0;
    end loop;

    -- initiate force
    force := new terms.fmat2 (1 .. nn, 1 .. 3);
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := 0.0;
      end loop;
    end loop;
  end setup;

  procedure autocad is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    infinity : constant long_float := 1.0E+20;

    glob           : integer;
    lo,le,xc,yc,zc : long_float;
    vx             : fmat3x1;
    io             : file_type;

  begin
    create(io, out_file, "autocad.scr");

    -- set running object snap mode to none
    put(io,"osmode 0");
    new_line(io);

    -- undeformed cable elements
    if cable'last > 0 then
      put(io,"layer make Undeformed_Cable ");
      new_line(io);

      for e in cable'range loop
        put(io,"3dpoly ");
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          put(io,xo(glob),1,8,3);
          put(io,",");
          put(io,yo(glob),1,8,3);
          put(io,",");
          put(io,zo(glob),1,8,3);
          put(io," ");
        end loop;
        new_line(io);
      end loop;
    end if;

    -- undeformed membrane elements
    if membr'last > 0 then
      put(io,"layer make Undeformed_Membrane ");
      new_line(io);

      -- create a polygon mesh
      put(io,"pface");
      new_line(io);

      -- vertex location
      for glob in 1 .. nn loop
        put(io,xo(glob),1,8,3);
        put(io,",");
        put(io,yo(glob),1,8,3);
        put(io,",");
        put(io,zo(glob),1,8,3);
        new_line(io);
      end loop;
      new_line(io);

      -- define faces
      for e in membr'range loop
        put(io,membr(e).glob(1),0);
        put(io," ");
        put(io,membr(e).glob(2),0);
        put(io," ");
        put(io,membr(e).glob(3),0);
        new_line(io);
        new_line(io);
      end loop;
      new_line(io);
    end if;

    -- text style
    put(io,"textstyle Standard");
    new_line(io);

    -- text size
    lo := infinity;

    -- cable elements
    if cable'last > 0 then
      le := 0.0;
      for e in cable'range loop
        le := le + cable(e).length;
      end loop;

      le := le / long_float(cable'last);

      if le < lo then
        lo := le;
      end if;
    end if;

    -- membrane elements
    if membr'last > 0 then
      le := 0.0;
      for e in membr'range loop
        le := le + membr(e).area;
      end loop;

      le := 1.519671 * sqrt(le / long_float(membr'last));

      if le < lo then
        lo := le;
      end if;
    end if;

    -- text size
    put(io,"textsize ");
    put(io,(0.1 * lo),1,8,3);
    new_line(io);

    -- node numbers
    put(io,"layer make Node_Numbers ");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,"text BL ");

      put(io,xo(glob),1,8,3);
      put(io,",");
      put(io,yo(glob),1,8,3);
      put(io,",");
      put(io,zo(glob),1,8,3);

      put(io,"   ");
      put(io,glob,0);
      put(io," ");

      new_line(io);
    end loop;

    -- cable numbers
    if cable'last > 0 then
      put(io,"layer make Cable_Numbers ");
      new_line(io);

      for e in cable'range loop
        xc := 0.0;
        yc := 0.0;
        zc := 0.0;
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          xc := xc + xo(glob);
          yc := yc + yo(glob);
          zc := zc + zo(glob);
        end loop;
        xc := xc / 2.0;
        yc := yc / 2.0;
        zc := zc / 2.0;

        put(io,"text MC ");

        put(io,xc,1,8,3);
        put(io,",");
        put(io,yc,1,8,3);
        put(io,",");
        put(io,zc,1,8,3);

        put(io,"   ");
        put(io,e,0);
        put(io," ");

        new_line(io);
      end loop;
    end if;

    -- membrane numbers
    if membr'last > 0 then
      put(io,"layer make Membrane_Numbers ");
      new_line(io);

      for e in membr'range loop
        xc := 0.0;
        yc := 0.0;
        zc := 0.0;
        for loc in 1 .. 3 loop
          glob := membr(e).glob(loc);
          xc := xc + xo(glob);
          yc := yc + yo(glob);
          zc := zc + zo(glob);
        end loop;
        xc := xc / 3.0;
        yc := yc / 3.0;
        zc := zc / 3.0;

        put(io,"text MC ");

        put(io,xc,1,8,3);
        put(io,",");
        put(io,yc,1,8,3);
        put(io,",");
        put(io,zc,1,8,3);

        put(io,"   ");
        put(io,e,0);
        put(io," ");

        new_line(io);
      end loop;
    end if;

    -- deformed cable elements
    if cable'last > 0 then
      put(io,"layer make Deformed_Cable ");
      new_line(io);

      for e in cable'range loop
        put(io,"3dpoly ");
        for loc in 1 .. 2 loop
          glob := cable(e).glob(loc);
          vertex(glob,vx);
          put(io,vx(1),1,8,3);
          put(io,",");
          put(io,vx(2),1,8,3);
          put(io,",");
          put(io,vx(3),1,8,3);
          put(io," ");
        end loop;
        new_line(io);
      end loop;
    end if;

    -- deformed membrane elements
    if membr'last > 0 then
      put(io,"layer make Deformed_Membrane ");
      new_line(io);

      -- create a polygon mesh
      put(io,"pface");
      new_line(io);

      -- vertex location
      for glob in 1 .. nn loop
        vertex(glob,vx);
        put(io,vx(1),1,8,3);
        put(io,",");
        put(io,vx(2),1,8,3);
        put(io,",");
        put(io,vx(3),1,8,3);
        new_line(io);
      end loop;
      new_line(io);

      -- define faces
      for e in membr'range loop
        put(io,membr(e).glob(1),0);
        put(io," ");
        put(io,membr(e).glob(2),0);
        put(io," ");
        put(io,membr(e).glob(3),0);
        new_line(io);
        new_line(io);
      end loop;
      new_line(io);
    end if;

    put(io,"layer set 0 ");
    new_line(io);

    if cable'last > 0 then
      put(io,"layer off Undeformed_Cable ");
      new_line(io);
      put(io,"layer off Cable_Numbers ");
      new_line(io);
    end if;

    if membr'last > 0 then
      put(io,"layer off Undeformed_Membrane ");
      new_line(io);
      put(io,"layer off Membrane_Numbers ");
      new_line(io);
    end if;

    put(io,"layer off Node_Numbers ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad;

  procedure output is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    glob,i,last,s               : integer;
    strain,axial,st1,st2,st3,fe : long_float;
    s1,s2,cos1,cos2,cos3,aux    : long_float;
    vx                          : fmat3x1;
    gc                          : fmat2x3;
    gm                          : fmat9x1;
    io                          : file_type;

  begin
    create(io, out_file, "output.txt");

    putchar(io,'-',65);
    new_line(io);

    put(io,"| Major Iterations                            | ");
    put(io,major,15);
    put(io," |");
    new_line(io);

    put(io,"| Last Minor Iterations                       | ");
    put(io,minor,15);
    put(io," |");
    new_line(io);

    put(io,"| Total Potential Energy                      | ");
    put(io,f,2,8,3);
    put(io," |");
    new_line(io);

    put(io,"| Allowable Maximum Residue                   | ");
    put(io,maxnorm,2,8,3);
    put(io," |");
    new_line(io);

    putchar(io,'-',65);
    new_line(io);

    -- displacements
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Displ X             Displ Y             Displ Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        i := index(glob,axis);
        if i > n then
          put(io,known(glob,axis),2,8,3);
        else
          put(io,x(i),2,8,3);
        end if;
      end loop;
      new_line(io);
    end loop;

    -- vertex location
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Coord X             Coord Y             Coord Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,vx);
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        put(io,vx(axis),2,8,3);
      end loop;
      new_line(io);
    end loop;

    -- initiate support reactions
    for glob in 1 .. nn loop
      for axis in 1 .. 3 loop
        force(glob,axis) := -force(glob,axis);
      end loop;
    end loop;

    -- cable elements
    new_line(io);
    new_line(io);
    putchar(io,'-',25);
    new_line(io);
    put(io," Elem               Force");
    new_line(io);
    putchar(io,'-',25);
    new_line(io);

    new_line(io);

    for e in cable'range loop
      cableobj(cable(e),strain,axial,fe,x,gc);

      -- axial forces
      put(io,e,5);
      putchar(io,' ',5);
      put(io,axial,2,8,3);
      new_line(io);

      -- update support reactions
      for loc in 1 .. 2 loop
        glob := cable(e).glob(loc);
        for axis in 1 .. 3 loop
          force(glob,axis) := force(glob,axis) + gc(loc,axis);
        end loop;
      end loop;
    end loop;

    -- membrane elements
    new_line(io);
    new_line(io);
    putchar(io,'-',45);
    new_line(io);
    put(io," Elem            Stress-1            Stress-2");
    new_line(io);
    putchar(io,'-',45);
    new_line(io);

    new_line(io);

    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gm);

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - membr(e).u2(axis) * membr(e).u3(axis);
      end loop;

      if cos1 > 1.0 then
        cos1 := 1.0;
      end if;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - membr(e).u3(axis) * membr(e).u1(axis);
      end loop;

      if cos2 > 1.0 then
        cos2 := 1.0;
      end if;

      -- cosine alpha3
      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 - membr(e).u1(axis) * membr(e).u2(axis);
      end loop;

      if cos3 > 1.0 then
        cos3 := 1.0;
      end if;

      -- principal stresses
      aux := 0.0;
      aux := aux + st1 * st2 * (cos3 ** 2 - 0.5);
      aux := aux + st1 * st3 * (cos2 ** 2 - 0.5);
      aux := aux + st2 * st3 * (cos1 ** 2 - 0.5);
      aux := 4.0 * aux + st1 ** 2 + st2 ** 2 + st3 ** 2;
      aux := sqrt(aux);

      s1 := ((st1 + st2 + st3) + aux) / 2.0;
      s2 := ((st1 + st2 + st3) - aux) / 2.0;

      put(io,e,5);
      putchar(io,' ',5);
      put(io,s1,2,8,3);
      putchar(io,' ',5);
      put(io,s2,2,8,3);
      new_line(io);

      -- update support reactions
      s := 0;
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        for axis in 1 .. 3 loop
          s := s + 1;
          force(glob,axis) := force(glob,axis) + gm(s);
        end loop;
      end loop;
    end loop;

    -- support reactions
    new_line(io);
    new_line(io);
    putchar(io,'-',65);
    new_line(io);
    put(io," Node             Force X             Force Y             Force Z");
    new_line(io);
    putchar(io,'-',65);
    new_line(io);

    new_line(io);

    for glob in 1 .. nn loop
      last := 0;
      for axis in 1 .. 3 loop
        if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
          last := axis;
        end if;
      end loop;

      if last /= 0 then
        put(io,glob,5);
        for axis in 1 .. last loop
          if index(glob,axis) > n and abs force(glob,axis) > maxnorm then
            putchar(io,' ',5);
            put(io,force(glob,axis),2,8,3);
          else
            putchar(io,' ',20);
          end if;
        end loop;
        new_line(io);
      end if;
    end loop;

    close(io);
  end output;

  procedure gid is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics.long_elementary_functions;

    glob,i               : integer;
    aux,st1,st2,st3,fe   : long_float;
    s1,s2,cos1,cos2,cos3 : long_float;
    gm                   : fmat9x1;
    adjacent             : terms.imat1_ptr;
    stress1,stress2      : terms.fmat1_ptr;
    vx                   : fmat3x1;
    io                   : file_type;

  begin
    create(io, out_file, "mesh.gid");

    -- deformed mesh
    put(io,"Mesh ""Deformed"" Dimension 3 ElemType Triangle NNode 3");
    new_line(io);

    new_line(io);

    -- coordinates
    put(io,"Coordinates");
    new_line(io);

    for glob in 1 .. nn loop
      vertex(glob,vx);
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        put(io,vx(axis),2,8,3);
      end loop;
      new_line(io);
    end loop;

    put(io,"End Coordinates");
    new_line(io);

    new_line(io);

    -- elements
    put(io,"Elements");
    new_line(io);

    for e in membr'range loop
      put(io,e,5);
      putchar(io,' ',5);
      put(io,membr(e).glob(1),5);
      putchar(io,' ',5);
      put(io,membr(e).glob(2),5);
      putchar(io,' ',5);
      put(io,membr(e).glob(3),5);
      new_line(io);
    end loop;

    put(io,"End elements");
    new_line(io);

    close(io);

    -- Calculate the principal stress values for each element. For
    -- each node, average the values from the elements adjacent to
    -- this node.

    -- initiate adjacent, stress1, stress2
    adjacent := new terms.imat1 (1 .. nn);
    stress1 := new terms.fmat1 (1 .. nn);
    stress2 := new terms.fmat1 (1 .. nn);

    for glob in 1 .. nn loop
      adjacent(glob) := 0;
      stress1(glob) := 0.0;
      stress2(glob) := 0.0;
    end loop;

    -- principal stress
    for e in membr'range loop
      membrobj(membr(e),st1,st2,st3,fe,x,gm);

      -- cosine alpha1
      cos1 := 0.0;
      for axis in 1 .. 3 loop
        cos1 := cos1 - membr(e).u2(axis) * membr(e).u3(axis);
      end loop;

      if cos1 > 1.0 then
        cos1 := 1.0;
      end if;

      -- cosine alpha2
      cos2 := 0.0;
      for axis in 1 .. 3 loop
        cos2 := cos2 - membr(e).u3(axis) * membr(e).u1(axis);
      end loop;

      if cos2 > 1.0 then
        cos2 := 1.0;
      end if;

      -- cosine alpha3
      cos3 := 0.0;
      for axis in 1 .. 3 loop
        cos3 := cos3 - membr(e).u1(axis) * membr(e).u2(axis);
      end loop;

      if cos3 > 1.0 then
        cos3 := 1.0;
      end if;

      -- principal stresses
      aux := 0.0;
      aux := aux + st1 * st2 * (cos3 ** 2 - 0.5);
      aux := aux + st1 * st3 * (cos2 ** 2 - 0.5);
      aux := aux + st2 * st3 * (cos1 ** 2 - 0.5);
      aux := 4.0 * aux + st1 ** 2 + st2 ** 2 + st3 ** 2;
      aux := sqrt(aux);

      s1 := ((st1 + st2 + st3) + aux) / 2.0;
      s2 := ((st1 + st2 + st3) - aux) / 2.0;

      -- update adjacent, stress1, stress2
      for loc in 1 .. 3 loop
        glob := membr(e).glob(loc);
        adjacent(glob) := adjacent(glob) + 1;
        stress1(glob) := stress1(glob) + s1;
        stress2(glob) := stress2(glob) + s2;
      end loop;
    end loop;

    -- average
    for glob in 1 .. nn loop
      stress1(glob) := stress1(glob) / long_float(adjacent(glob));
      stress2(glob) := stress2(glob) / long_float(adjacent(glob));
    end loop;

    create(io, out_file, "results.gid");

    put(io,"Gid Post Results File 1.0");
    new_line(io);

    -- displacement
    new_line(io);

    put(io,"Result ""Displacement"" ""Analysis"" 1 Vector OnNodes");
    new_line(io);
    put(io,"ComponentNames ""X Displ"", ""Y Displ"", ""Z Displ""");
    new_line(io);
    put(io,"Values");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      for axis in 1 .. 3 loop
        putchar(io,' ',5);
        i := index(glob,axis);
        if i > n then
          put(io,known(glob,axis),2,8,3);
        else
          put(io,x(i),2,8,3);
        end if;
      end loop;
      new_line(io);
    end loop;

    put(io,"End Values");
    new_line(io);

    -- principal stress
    new_line(io);

    put(io,"Result ""Stress 1"" ""Analysis"" 1 Scalar OnNodes");
    new_line(io);
    put(io,"Values");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      putchar(io,' ',5);
      put(io,stress1(glob),2,8,3);
      new_line(io);
    end loop;

    put(io,"End Values");
    new_line(io);

    -- principal stress
    new_line(io);

    put(io,"Result ""Stress 2"" ""Analysis"" 1 Scalar OnNodes");
    new_line(io);
    put(io,"Values");
    new_line(io);

    for glob in 1 .. nn loop
      put(io,glob,5);
      putchar(io,' ',5);
      put(io,stress2(glob),2,8,3);
      new_line(io);
    end loop;

    put(io,"End Values");
    new_line(io);

    close(io);
  end gid;

begin
  input;

  setup;

  setforce;

  objective(n,f,x,g);

  -- mntol
  if mntol <= 0.0 then
    mntol := 0.001;
  end if;

  -- mnref
  if mnref <= 0.0 then
    mnref := nlp.norm(n,g);
  end if;

  -- maxnorm
  maxnorm := mntol * mnref;

  -- maxminor
  if maxminor <= 0 then
    maxminor := 10 * n;
  end if;

  major := 0;
  while major < maxmajor and nlp.norm(n,g) > maxnorm loop
    major := major + 1;

    qn.lbfgs(maxlist,n,maxminor,maxls,
             minor,
             maxnorm,lstol,
             f,
             x,g,
             optimal);

    setforce;

    objective(n,f,x,g);
  end loop;

  if optimal then
    autocad;
    output;
    gid;
  end if;
end formfind;
